@@ -1,5 +1,5 @@
 package PAR;
-$PAR::VERSION = '0.970_01';
+$PAR::VERSION = '0.981_01';
 
 use 5.006;
 use strict;
@@ -7,13 +7,35 @@ use warnings;
 use Config '%Config';
 use Carp qw/croak/;
 
+# If the 'prefork' module is available, we
+# register various run-time loaded modules with it.
+# That way, there is more shared memory in a forking
+# environment.
+BEGIN {
+    if (eval 'require prefork') {
+        prefork->import($_) for qw/
+            Archive::Zip
+            File::Glob
+            File::Spec
+            File::Temp
+            LWP::Simple
+            PAR::Heavy
+        /;
+        # not including Archive::Unzip::Burst which only makes sense
+        # in the context of a PAR::Packer'ed executable anyway.
+    }
+}
+
+use PAR::SetupProgname;
+use PAR::SetupTemp;
+
 =head1 NAME
 
 PAR - Perl Archive Toolkit
 
 =head1 VERSION
 
-This document describes version 0.971 of PAR, released January 10, 2007.
+This document describes development release 0.981_01 of PAR, released August 10, 2008.
 
 =head1 SYNOPSIS
 
@@ -98,7 +120,7 @@ F<script/myscript> exits, so does your main program. To make this more useful,
 you can defer this to runtime: (otherwise equivalent)
 
     require PAR;
-    PAR->import( { file => 'foo.par', run => 'myscript' };
+    PAR->import( { file => 'foo.par', run => 'myscript' } );
 
 If you have L<PAR::Repository::Client> installed, you can do this:
 
@@ -233,6 +255,47 @@ By default, PAR strips POD sections from bundled modules. In case
 that causes trouble, you can turn this off by setting the
 environment variable C<PAR_VERBATIM> to C<1>.
 
+=head2 import options
+
+When you "use PAR {...}" or call PAR->import({...}), the following
+options are available.
+
+  PAR->import({ file => 'foo.par' });
+  # or
+  PAR->import({ repository => 'http://foo/bar/' });
+
+=over
+
+=item file
+
+The par filename.
+
+You must pass I<one> option of either 'file' or 'repository'.
+
+=item repository
+
+A par repository (exclusive of file)
+
+=item fallback
+
+Search the system @INC before the par.
+
+=item run
+
+The name of a script to run in the par.  Exits when done.
+
+=item no_shlib_unpack
+
+Skip unpacking bundled dynamic libraries from shlib/$archname.  The
+client may have them installed, or you may wish to cache them yourself.
+In either case, they must end up in the standard install location (such
+as /usr/local/lib/) or in $ENV{PAR_TEMP} I<before> you require the
+module which needs them.  If they are not accessible before you require
+the dependent module, perl will die with a message such as "cannot open
+shared object file..." 
+
+=back
+
 =cut
 
 use vars qw(@PAR_INC);              # explicitly stated PAR library files (prefered)
@@ -242,6 +305,9 @@ use vars qw(@LibCache %LibCache);   # I really miss pseudohash.
 use vars qw($LastAccessedPAR $LastTempFile);
 use vars qw(@RepositoryObjects);    # If we have PAR::Repository::Client support, we
                                     # put the ::Client objects in here.
+use vars qw(%FileCache);            # The Zip-file file-name-cache
+                                    # Layout:
+                                    # $FileCache{$ZipObj}{$FileName} = $Member
 
 my $ver  = $Config{version};
 my $arch = $Config{archname};
@@ -251,29 +317,16 @@ my $is_insensitive_fs = (
         and (-s lc($progname) || -1) == (-s uc($progname) || -1)
         and (-s lc($progname) || -1) == -s $progname
 );
-my $par_temp;
-
-
-# The following section overrides a bug in AutoLoader which results in
-# infinite recursion in case of AUTOLOAD and re-named files.
-# See http://www.nntp.perl.org/group/perl.par/2796 et al.
-require AutoLoader;
-if ($AutoLoader::VERSION <= 5.61) {
-    no strict;
-    no warnings;
-    # If the following doesn't scare you then you scare me.
-    delete $INC{"AutoLoader.pm"};
-    require PAR::AutoLoaderFix;
-    $INC{"AutoLoader.pm"} = $INC{"PAR/AutoLoaderFix.pm"};
-}
 
+# lexical for import(), and _import_foo() functions to control unpar()
+my %unpar_options;
 
 # called on "use PAR"
 sub import {
     my $class = shift;
 
-    _set_progname();
-    _set_par_temp();
+    PAR::SetupProgname::set_progname();
+    PAR::SetupTemp::set_par_temp_env();
 
     $progname = $ENV{PAR_PROGNAME} ||= $0;
     $is_insensitive_fs = (-s $progname and (-s lc($progname) || -1) == (-s uc($progname) || -1));
@@ -360,6 +413,10 @@ sub import {
 # import() helper for the "use PAR {...};" syntax.
 sub _import_hash_ref {
     my $opt = shift;
+
+    # hash slice assignment -- pass all of the options into unpar
+    local @unpar_options{keys(%$opt)} = values(%$opt);
+
     # check for incompatible options:
     if ( exists $opt->{repository} and exists $opt->{file} ) {
         croak("Invalid PAR loading options. Cannot have a 'repository' and 'file' option at the same time.");
@@ -466,14 +523,15 @@ sub _import_repository {
     return $obj;
 }
 
+# Given an Archive::Zip obj and a list of files/paths,
+# this function returns the Archive::Zip::Member for the
+# first of the files found in the ZIP. If none is found,
+# returns the empty list.
 sub _first_member {
     my $zip = shift;
-    my %names = map { ( $_->fileName => $_ ) } $zip->members;
-    my %lc_names;
-    %lc_names = map { ( lc($_->fileName) => $_ ) } $zip->members if $is_insensitive_fs;
     foreach my $name (@_) {
-        return $names{$name} if $names{$name};
-        return $lc_names{lc($name)} if $is_insensitive_fs and $lc_names{lc($name)};
+        my $member = _cached_member_named($zip, $name);
+        return $member if $member;
     }
     return;
 }
@@ -506,28 +564,41 @@ sub _run_member {
 
 sub _extract_inc {
     my $file = shift;
-    my $inc = "$par_temp/inc";
-    my $dlext = do {
-        require Config;
-        (defined %Config::Config) ? $Config::Config{dlext} : '';
-    };
+    my $inc = "$PAR::SetupTemp::PARTemp/inc";
+    my $dlext = defined($Config{dlext}) ? $Config::Config{dlext} : '';
 
     if (!-d $inc) {
         for (1 .. 10) { mkdir("$inc.lock", 0755) and last; sleep 1 }
+        
+        # First try to unzip the *fast* way.
+        eval {
+          require Archive::Unzip::Burst;
+          Archive::Unzip::Burst::unzip($file, $inc)
+            and die "Could not unzip '$file' into '$inc'. Error: $!";
+        };
 
-        open my $fh, '<', $file or die "Cannot find '$file': $!";
-        binmode($fh);
-        bless($fh, 'IO::File');
-
-        my $zip = Archive::Zip->new;
-        ( $zip->readFromFileHandle($fh, $file) == Archive::Zip::AZ_OK() )
-            or die "Read '$file' error: $!";
-
-        for ( $zip->memberNames() ) {
-            next if m{\.\Q$dlext\E[^/]*$};
-            s{^/}{};
-            $zip->extractMember($_, "$inc/" . $_);
+        # This means the fast module is there, but didn't work.
+        if ($@ =~ /^Could not unzip/) {
+          die $@;
         }
+
+        # failed to load Archive::Unzip::Burst. Default to slow way.
+        elsif ($@) {
+          open my $fh, '<', $file or die "Cannot find '$file': $!";
+          binmode($fh);
+          bless($fh, 'IO::File');
+
+          my $zip = Archive::Zip->new;
+          ( $zip->readFromFileHandle($fh, $file) == Archive::Zip::AZ_OK() )
+              or die "Read '$file' error: $!";
+
+          for ( $zip->memberNames() ) {
+              next if m{\.\Q$dlext\E[^/]*$};
+              s{^/}{};
+              $zip->extractMember($_, "$inc/" . $_);
+          }
+        }
+        
         rmdir("$inc.lock");
     }
 
@@ -610,6 +681,7 @@ sub reload_libs {
     foreach my $par (@par_files) {
         my $inc_ref = $PAR_INC{$par} or next;
         delete $LibCache{$par};
+        delete $FileCache{$par};
         foreach my $file (sort keys %$inc_ref) {
             delete $INC{$file};
             require $file;
@@ -617,6 +689,17 @@ sub reload_libs {
     }
 }
 
+#sub find_zip_member {
+#    my $file = pop;
+#
+#    foreach my $zip (@LibCache) {
+#        my $member = _first_member($zip, $file) or next;
+#        return $member;
+#    }
+#
+#    return;
+#}
+
 sub read_file {
     my $file = pop;
 
@@ -640,10 +723,12 @@ sub unpar {
     my $zip = $LibCache{$par};
     my @rv = $par;
 
+    # a guard against (currently unimplemented) recursion
     return if $PAR::__unpar;
     local $PAR::__unpar = 1;
 
     unless ($zip) {
+        # URL use case ==> download
         if ($par =~ m!^\w+://!) {
             require File::Spec;
             require LWP::Simple;
@@ -651,6 +736,7 @@ sub unpar {
             # reflector support
             $par .= "pm=$file" if $par =~ /[?&;]/;
 
+            # prepare cache directory
             $ENV{PAR_CACHE} ||= '_par';
             mkdir $ENV{PAR_CACHE}, 0777;
             if (!-d $ENV{PAR_CACHE}) {
@@ -659,6 +745,12 @@ sub unpar {
                 return unless -d $ENV{PAR_CACHE};
             }
 
+            # Munge URL into local file name
+            # FIXME: This might result in unbelievably long file names!
+            # I have run into the file/path length limitations of linux
+            # with similar code in PAR::Repository::Client.
+            # I suspect this is even worse on Win32.
+            # -- Steffen
             my $file = $par;
             if (!%escapes) {
                 $escapes{chr($_)} = sprintf("%%%02X", $_) for 0..255;
@@ -667,16 +759,20 @@ sub unpar {
                 use bytes;
                 $file =~ s/([^\w\.])/$escapes{$1}/g;
             }
+
             $file = File::Spec->catfile( $ENV{PAR_CACHE}, $file);
             LWP::Simple::mirror( $par, $file );
-            return unless -e $file;
+            return unless -e $file and -f _;
             $par = $file;
         }
+        # Got the .par as a string. (reference to scalar, of course)
         elsif (ref($par) eq 'SCALAR') {
             my ($fh) = _tempfile();
             print $fh $$par;
             $par = $fh;
         }
+        # If the par is not a valid .par file name and we're being strict
+        # about this, then also check whether "$par.par" exists
         elsif (!(($allow_other_ext or $par =~ /\.par\z/i) and -f $par)) {
             $par .= ".par";
             return unless -f $par;
@@ -685,9 +781,9 @@ sub unpar {
         require Archive::Zip;
         $zip = Archive::Zip->new;
 
-    	my @file;
+        my @file;
         if (!ref $par) {
-	        @file = $par;
+            @file = $par;
 
             open my $fh, '<', $par;
             binmode($fh);
@@ -703,8 +799,10 @@ sub unpar {
 
         push @LibCache, $zip;
         $LibCache{$_[0]} = $zip;
+        $FileCache{$_[0]} = _make_file_cache($zip);
 
-        foreach my $member ( $zip->membersMatching(
+        # only recursive case -- appears to be unused and unimplemented
+        foreach my $member ( _cached_members_matching($zip, 
             "^par/(?:$Config{version}/)?(?:$Config{archname}/)?"
         ) ) {
             next if $member->isDirectory;
@@ -713,24 +811,32 @@ sub unpar {
             push @rv, unpar(\$content, undef, undef, 1);
         }
         
-        # extract all dlls from the .par to $ENV{PAR_TEMP}!
-        # XXX is this correct?
+        # extract all shlib dlls from the .par to $ENV{PAR_TEMP}
         # Intended to fix problem with Alien::wxWidgets/Wx...
-        # for all zip members that end in .dll/.so (depending on platform)
-        foreach my $member ( $zip->membersMatching(
-            '\.'.quotemeta($Config{dlext}).'$'
-        ) ) {
-            next if $member->isDirectory or !$ENV{PAR_TEMP};
-            my $member_name = $member->fileName;
-            next unless $member_name =~ m{
-                    \/([^/]+)$
-                }x
-                or $member_name =~ m{
-                    ^([^/]+)$
-                };
-            my $extract_name = $1;
-            my $dest_name = File::Spec->catfile($ENV{PAR_TEMP}, $extract_name);
-            $member->extractToFileNamed($dest_name);
+        # NOTE auto/foo/foo.so|dll will get handled by the dynaloader
+        # hook, so no need to pull it out here.
+        # Allow this to be disabled so caller can do their own caching
+        # via import({no_shlib_unpack => 1, file => foo.par})
+        if(not $unpar_options{no_shlib_unpack} and defined $ENV{PAR_TEMP}) {
+            my @members = _cached_members_matching( $zip,
+              qr#^shlib/$Config{archname}/.*\.\Q$Config{dlext}\E(?:\.|$)#
+            );
+            foreach my $member (@members) {
+                next if $member->isDirectory;
+                my $member_name = $member->fileName;
+                next unless $member_name =~ m{
+                        \/([^/]+)$
+                    }x
+                    or $member_name =~ m{
+                        ^([^/]+)$
+                    };
+                my $extract_name = $1;
+                my $dest_name =
+                    File::Spec->catfile($ENV{PAR_TEMP}, $extract_name);
+                # but don't extract it if we've already got one
+                $member->extractToFileNamed($dest_name)
+                    unless(-e $dest_name);
+            }
         }
 
         # Now push this path into usual library search paths
@@ -784,75 +890,6 @@ sub unpar {
     return $fh;
 }
 
-# The C version of this code appears in myldr/mktmpdir.c
-sub _set_par_temp {
-    if ($ENV{PAR_TEMP} and $ENV{PAR_TEMP} =~ /(.+)/) {
-        $par_temp = $1;
-        return;
-    }
-
-    require File::Spec;
-
-    foreach my $path (
-        (map $ENV{$_}, qw( PAR_TMPDIR TMPDIR TEMPDIR TEMP TMP )),
-        qw( C:\\TEMP /tmp . )
-    ) {
-        next unless $path and -d $path and -w $path;
-        my $username;
-        my $pwuid;
-        # does not work everywhere:
-        eval {($pwuid) = getpwuid($>) if defined $>;};
-
-        if ( defined(&Win32::LoginName) ) {
-            $username = &Win32::LoginName;
-        }
-        elsif (defined $pwuid) {
-            $username = $pwuid;
-        }
-        else {
-            $username = $ENV{USERNAME} || $ENV{USER} || 'SYSTEM';
-        }
-        $username =~ s/\W/_/g;
-
-        my $stmpdir = File::Spec->catdir($path, "par-$username");
-        ($stmpdir) = $stmpdir =~ /^(.*)$/s;
-        mkdir $stmpdir, 0755;
-        if (!$ENV{PAR_CLEAN} and my $mtime = (stat($progname))[9]) {
-            my $ctx = eval { require Digest::SHA; Digest::SHA->new(1) }
-                   || eval { require Digest::SHA1; Digest::SHA1->new }
-                   || eval { require Digest::MD5; Digest::MD5->new };
-
-            # Workaround for bug in Digest::SHA 5.38 and 5.39
-            my $sha_version = eval { $Digest::SHA::VERSION } || 0;
-            if ($sha_version eq '5.38' or $sha_version eq '5.39') {
-                $ctx->addfile($progname, "b") if ($ctx);
-            }
-            else {
-                if ($ctx and open(my $fh, "<$progname")) {
-                    binmode($fh);
-                    $ctx->addfile($fh);
-                    close($fh);
-                }
-            }
-
-            $stmpdir = File::Spec->catdir(
-                $stmpdir,
-                "cache-" . ( $ctx ? $ctx->hexdigest : $mtime )
-            );
-        }
-        else {
-            $ENV{PAR_CLEAN} = 1;
-            $stmpdir = File::Spec->catdir($stmpdir, "temp-$$");
-        }
-
-        $ENV{PAR_TEMP} = $stmpdir;
-        mkdir $stmpdir, 0755;
-        last;
-    }
-
-    $par_temp = $1 if $ENV{PAR_TEMP} and $ENV{PAR_TEMP} =~ /(.+)/;
-}
-
 sub _tempfile {
     my ($fh, $filename);
     if ($ENV{PAR_CLEAN} or !@_) {
@@ -863,8 +900,8 @@ sub _tempfile {
             # and will be deleted by the C runtime; having File::Temp
             # delete it has the only effect of giving ugly warnings
             ($fh, $filename) = File::Temp::tempfile(
-                DIR     => $par_temp,
-                UNLINK  => ($^O ne 'MSWin32'),
+                DIR     => $PAR::SetupTemp::PARTemp,
+                UNLINK  => ($^O ne 'MSWin32' and $^O !~ /hpux/),
             ) or die "Cannot create temporary file: $!";
             binmode($fh);
             return ($fh, 1, $filename);
@@ -874,7 +911,7 @@ sub _tempfile {
     require File::Spec;
 
     # untainting tempfile path
-    local $_ = File::Spec->catfile( $par_temp, $_[0] );
+    local $_ = File::Spec->catfile( $PAR::SetupTemp::PARTemp, $_[0] );
     /^(.+)$/ and $filename = $1;
 
     if (-r $filename) {
@@ -888,33 +925,54 @@ sub _tempfile {
     return ($fh, 1, $filename);
 }
 
-sub _set_progname {
-    require File::Spec;
-
-    if ($ENV{PAR_PROGNAME} and $ENV{PAR_PROGNAME} =~ /(.+)/) {
-        $progname = $1;
+# Given an Archive::Zip object, this generates a hash of
+#   file_name_in_zip => file object
+# and returns a reference to that.
+# If we broke the encapsulation of A::Zip::Member and
+# accessed $member->{fileName} directly, that would be
+# *significantly* faster.
+sub _make_file_cache {
+    my $zip = shift;
+    if (not ref($zip)) {
+        croak("_make_file_cache needs an Archive::Zip object as argument.");
     }
-    $progname ||= $0;
-
-    if (( () = File::Spec->splitdir($progname) ) > 1 or !$ENV{PAR_PROGNAME}) {
-        if (open my $fh, $progname) {
-            return if -s $fh;
-        }
-        if (-s "$progname$Config{_exe}") {
-            $progname .= $Config{_exe};
-            return;
-        }
-    }
-
-    foreach my $dir (split /\Q$Config{path_sep}\E/, $ENV{PATH}) {
-        next if exists $ENV{PAR_TEMP} and $dir eq $ENV{PAR_TEMP};
-        my $name = File::Spec->catfile($dir, "$progname$Config{_exe}");
-        if (-s $name) { $progname = $name; last }
-        $name = File::Spec->catfile($dir, "$progname");
-        if (-s $name) { $progname = $name; last }
+    my $cache = {};
+    foreach my $member ($zip->members) {
+        $cache->{$member->fileName()} = $member;
     }
+    return $cache;
 }
 
+# given an Archive::Zip object, this finds the cached hash
+# of Archive::Zip member names => members,
+# and returns all member objects whose file names match
+# a regexp
+# Without file caching, it just uses $zip->membersMatching
+sub _cached_members_matching {
+    my $zip = shift;
+    my $regex = shift;
+
+    my $cache = $FileCache{$zip};
+    $cache = $FileCache{$zip} = _make_file_cache($zip) if not $cache;
+
+    return map {$cache->{$_}}
+        grep { $_ =~ $regex }
+        keys %$cache;
+}
+
+# access named zip file member through cache. Fall
+# back to using Archive::Zip (slow)
+sub _cached_member_named {
+    my $zip = shift;
+    my $name = shift;
+
+    my $cache = $FileCache{$zip};
+    $cache = $FileCache{$zip} = _make_file_cache($zip) if not $cache;
+    return $cache->{$name};
+}
+
+
+
 
 1;
 
@@ -939,6 +997,10 @@ L<Archive::Zip>, L<perlfunc/require>
 
 L<ex::lib::zip>, L<Acme::use::strict::with::pride>
 
+PAR supports the L<prefork> module. It declares various run-time
+dependencies so you can use the L<prefork> module to get streamlined
+processes in a forking environment.
+
 =head1 ACKNOWLEDGMENTS
 
 Nicholas Clark for pointing out the mad source filter hook within the
@@ -970,7 +1032,7 @@ preferred.
 
 =head1 COPYRIGHT
 
-Copyright 2002, 2003, 2004, 2005, 2006 by Audrey Tang
+Copyright 2002-2008 by Audrey Tang
 E<lt>cpan@audreyt.orgE<gt>.
 
 This program is free software; you can redistribute it and/or modify it
